## Logistic function
logistic = function(x, z=400) {
  return(1 / (1 + 10^(-x / z)))
}

##  Official rating
# INPUT
# matches: matches data frame
# teams: team names
# OUTPUT
# r: official rating
official = function(matches, teams) {
  # number of teams
  n = length(teams)
  
  # number of matches
  m = dim(matches)[1]
  
  # number of days
  days = m / (n/2)
  
  # rating matrix
  r = rep(0, n)
  
  for (i in 1:m) {
    spread = matches[i,"score1"] - matches[i,"score2"]
    if (spread > 0) {
      r[matches[i,"team1"]] = r[matches[i,"team1"]] + 3
    }
    if (spread < 0) {
      r[matches[i,"team2"]] = r[matches[i,"team2"]] + 3
    }
    if (spread == 0) {
      r[matches[i,"team1"]] = r[matches[i,"team1"]] + 1
      r[matches[i,"team2"]] = r[matches[i,"team2"]] + 1
    }
  }
  return(list(r = r))
}

##  Elo
# INPUT
# matches: matches data frame
# teams: team names
# zeta: logistic parameter
# k: update factor
# scores: use scores of matches?
# hf: home-field advantage
# OUTPUT
# r: rating vector
# wins: number of victories
# foreseen: number of foreseen victories
# accuracy: percentage of foreseen victories
elo = function(matches, teams, z = 400, k = 25, scores=FALSE, hf=0) {
  # number of teams
  n = length(teams)
  
  # number of matches
  m = dim(matches)[1]
  
  # old rating vector
  rold = rep(0, n)
  
  # new rating vector
  rnew = rep(0, n)
  
  wins = 0
  foreseen = 0
  
  for (i in 1:m) {
    # compute prediction accuracy
    spread = matches[i,"score1"] - matches[i,"score2"] 
    if (spread > 0) {
      wins = wins + 1
      if ((rold[matches[i,"team1"]] + hf) > rold[matches[i,"team2"]]) {
        foreseen = foreseen + 1
      }
    }
    if (spread < 0) {
      wins = wins + 1
      if (rold[matches[i,"team2"]] > (rold[matches[i,"team1"]] + hf)) {
        foreseen = foreseen + 1
      }
    }
    
    # team 1
    # compute score
    if (scores == FALSE) {
      spread = matches[i,"score1"] - matches[i,"score2"]
      if (spread > 0) {
        score = 1
      }
      if (spread == 0) {
        score = 0.5
      }
      if (spread < 0) {
        score = 0
      }
    } else {
      score = (matches[i,"score1"] + 1) / (matches[i,"score1"] + matches[i,"score2"] + 2)
    }
    
    # compute update
    delta = rold[matches[i,"team1"]] - rold[matches[i,"team2"]]
    mu = 1 / (1 + 10^(-delta / z))
    update = k * (score - mu)
    
    # update rating
    rnew[matches[i,"team1"]] = rold[matches[i,"team1"]] + update
    
    # team 2
    # compute score
    if (scores == FALSE) {
      spread = matches[i,"score2"] - matches[i,"score1"]
      if (spread > 0) {
        score = 1
      }
      if (spread == 0) {
        score = 0.5
      }
      if (spread < 0) {
        score = 0
      }
    } else {
      score = (matches[i,"score2"] + 1) / (matches[i,"score1"] + matches[i,"score2"] + 2)
    }
    
    # compute update
    delta = rold[matches[i,"team2"]] - rold[matches[i,"team1"]]
    mu = 1 / (1 + 10^(-delta / z))
    update = k * (score - mu)
    
    # update new rating
    rnew[matches[i,"team2"]] = rold[matches[i,"team2"]] + update

    # update old ratings
    rold[matches[i,"team1"]] = rnew[matches[i,"team1"]]
    rold[matches[i,"team2"]] = rnew[matches[i,"team2"]]
  }
  return(list(r=rnew, foreseen = foreseen, wins = wins, accuracy = foreseen / wins))
}


##  Massey
# ASSUMPTION: graph of A is connected
# INPUT
# matches: matches data frame
# teams: team names
# OUTPUT
# r: Massey rating (r = r1 + r2; r = o + d)
# r1: opponents rating
# r2: spread rating
# o: offensive rating
# d: defensive rating
# A: match matrix
# g: match graph
massey = function(matches, teams) {
  # number of teams
  n = length(teams)
  
  # number of matches
  m = dim(matches)[1]
  
  # match matrix
  X = matrix(0, nrow=m, ncol=n, byrow=TRUE)
  
  # margin of victory vector
  y = rep(0, m)
  
  # points for vector
  f = rep(0, n)
  
  # points against vector
  a = rep(0, n)
  
  # populate match matrix
  for (i in 1:m) {
    if (matches[i,"score1"] >= matches[i,"score2"]) {
      X[i, matches[i,"team1"]] = 1;
      X[i, matches[i,"team2"]] = -1;
      y[i] = matches[i,"score1"] - matches[i,"score2"]
    } else {
      X[i, matches[i,"team2"]] = 1;
      X[i, matches[i,"team1"]] = -1;
      y[i] = matches[i,"score2"] - matches[i,"score1"]
    }
    f[matches[i,"team1"]] = f[matches[i,"team1"]] + matches[i,"score1"] 
    f[matches[i,"team2"]] = f[matches[i,"team2"]] + matches[i,"score2"] 
    a[matches[i,"team1"]] = a[matches[i,"team1"]] + matches[i,"score2"] 
    a[matches[i,"team2"]] = a[matches[i,"team2"]] + matches[i,"score1"] 
  }
  
  # compute Massey matrix
  M = t(X) %*% X
  # compute team spread vector (or p = f-a)
  p = t(X) %*% y
  # check assumptions
  D = diag(diag(M), n, n)
  A = D - M
  g = graph_from_adjacency_matrix(A, mode="undirected", weighted=TRUE)
  if (!is.connected(g)) {
    cat("The graph is not connected");
    return(NULL)
  }
  
  # perturbate the system
  M1 = M
  p1 = p
  M1[n, ] = rep(1, n)
  p1[n,1] = 0
  # compute ranking r
  r = solve(M1, p1)
  
  # compute partial rankings r1 and r2 (r = r1 + r2)
  D1 = solve(D)
  r1 = D1 %*% A %*% r
  r2 = D1 %*% p
  
  # compute offensive and defensive rankings (r = o + d)
  N = D + A
  q = (D %*% r) - f
  # check if graph of A is bipartite (we know that the graph of A is connected)
  bipa = bipartite_mapping(graph_from_adjacency_matrix(A, mode="undirected"))
  if (bipa$res == TRUE) {
    # if bipartite, perturbate the system
    v = bipa$type
    v[v == TRUE] = 1
    v[v == FALSE] = -1
    N1 = N
    q1 = q
    N1[n, ] = v
    q1[n,1] = 0
    d = solve(N1, q1)
  } else {
    d = solve(N, q)
  }
  o = r - d
  r = as.vector(r)
  r1 = as.vector(r1)
  r2 = as.vector(r2)
  o = as.vector(o)
  d = as.vector(d)
  names(r) =  teams
  names(r1) =  teams
  names(r2) =  teams
  names(o) =  teams
  names(d) =  teams
  V(g)$teams = teams
  
  return(list(r=r, r1=r1, r2=r2, o=o, d=d, A = A, g = g))
}

# Keener
# ASSUMPTION: graph of A is strongly connected
# INPUT
# matches: matches data frame
# teams: team names
# OUTPUT
# r: Keener rating 
# value = eigenvalue
# A: match matrix
# g: match graph
keener = function(matches, teams) {
  # number of teams
  n = length(teams)
  
  # number of matches
  m = dim(matches)[1]
  
  # match matrix
  A = matrix(0, nrow=n, ncol=n, byrow=TRUE)
  
  # populate match matrix
  for (i in 1:m) {
    p = matches[i,"team1"]
    q = matches[i,"team2"]
    s1 = matches[i, "score1"]
    s2 = matches[i, "score2"]
    A[p,q] = (s1 + 1) / (s1 + s2 + 2);
    A[q,p] = (s2 + 1) / (s1 + s2 + 2);
  }
  
  # check assumptions
  g = graph_from_adjacency_matrix(A, mode="directed", weighted=TRUE)
  if (!is.connected(g, mode="strong")) {
    cat("The graph is not strongly connected")
    return(NULL)
  }
  
  
  e = eigen(A)
  v = e$values[1]
  r = Re(e$vectors[,1])
  if (r[1] < 0) {
    r = -r
  }
  names(r) = teams
  V(g)$teams = teams
  return(list(r = r, value = v, A = A, g = g))
}  


# Knight
# ASSUMPTION: A has total support (or graph of A is regularizable)
# INPUT
# matches: matches data frame
# teams: team names
# t: precision
# type: "laplace" uses Laplace rule of succession to build A, everything else uses the original method
# OUTPUT
# o: offensive rating 
# d: defensive rating 
# r: overall rating 
# iter = number of iterations
# A: match matrix
# g: match graph
knight = function(matches, teams, t = 6, type = "laplace") {
  # number of teams
  n = length(teams)
  
  # number of matches
  m = dim(matches)[1]
  
  # match matrix
  A = matrix(0, nrow=n, ncol=n, byrow=TRUE)
  
  # populate match matrix
  if (type == "laplace") {
    for (i in 1:m) {
      p = matches[i,"team1"]
      q = matches[i,"team2"]
      s1 = matches[i, "score1"]
      s2 = matches[i, "score2"]
      A[q,p] = (s1 + 1) / (s1 + s2 + 2);
      A[p,q] = (s2 + 1) / (s1 + s2 + 2);
    }
  } else {
    for (i in 1:m) {
      p = matches[i,"team1"]
      q = matches[i,"team2"]
      s1 = matches[i, "score1"]
      s2 = matches[i, "score2"]
      if (s1 > 0) {
        A[q,p] = s1;
      }
      if (s2 > 0) {
        A[p,q] = s2;
      }
    }
  }
  
  # check assumptions
  g = graph_from_adjacency_matrix(A, mode="directed", weighted=TRUE)
  #if ("g is not regularizable") {
  #  cat("The graph is not regularizable")
  #  return(NULL)
  #}
  
  
  y0 = rep(1, n);
  diff = 1
  eps = 1/10^t;
  iter = 0;
  while (diff > eps) {
    x1 = as.vector((1/y0) %*% A);
    y1 = as.vector(A %*% (1/x1));
    diff = sum(abs(y1 - y0));
    y0 = y1;
    iter = iter + 1;
  } 
  r = x1 / y1;
  names(x1) = teams;
  names(y1) = teams;
  names(r) = teams;
  V(g)$teams = teams
  return(list(o = x1, d = y1, r = r, iter = iter, A = A, g = g))
}  

# check if undirected graph is regularizable
# INPUT: g = graph
# OUTPUT: w = weights; r = degree
regularify = function (g) {
  n = vcount(g)
  m = ecount(g)
  E = get.edges(g, E(g))
  B = matrix(0, nrow = n, ncol = m)
  # build incidence matrix
  for (i in 1:m) {
    B[E[i,1], i] = 1
    B[E[i,2], i] = 1
  }  
  # objective function
  obj = rep(0, m + 1)
  # constraint matrix
  con = cbind(B, rep(-1, n))
  # direction of constraints
  dir = rep("=", n)
  # right hand side terms
  rhs = -degree(g)
  # solve the LP problem
  s = lp("max", obj, con, dir, rhs)
  # get the solution
  if (s$status == 0)
    return(list(w = s$solution[1:m] + 1, r = s$solution[m+1])) 
  else 
    return(NULL)   
}

# FUNCTION
# getDataFrameFromFile
# IN
# fileName: file name (csv) to read with mathes data (es: Serie A 07-08)
# day: day of competition to analyze
# OUT
# data frame from csv file

getDataFrameFromFile <- function(fileName, day = 38) {
  fileToRead = paste("dataset_soccer/", fileName, ".csv", sep = "")
  df = read.csv(file=fileToRead, head=TRUE, sep=",", stringsAsFactors = F)
  rows = day * 10
  df = head(df, rows)
  return(df)
}

# FUNCTION
# getTeamsFromDataFrame
# IN
# dataFrame: data frame of matches
# OUT
# vector of unique teams

getTeamsFromDataFrame <- function(dataFrame) {
  home = dataFrame[,"HomeTeam"]
  away = dataFrame[,"AwayTeam"]
  teams = c(home, away)
  return(unique(teams))
}

# FUNCTION
# prepareFootballData
# IN
# dataFrame: data frame of matches
# OUT
# clean data frame ready to compute ratings

prepareFootballData <- function(dataFrame) {
  
  # get all team names from data frame
  teamsName = getTeamsFromDataFrame(dataFrame)
  teams = seq(1, length(teamsName), 1)
  names(teams) = teamsName
  
  # assign unique id to teams, no string name
  for (i in 1:dim(dataFrame)[1]) {
    dataFrame[i,"HomeTeam"] = teams[dataFrame[i,"HomeTeam"]]
    dataFrame[i,"AwayTeam"] = teams[dataFrame[i,"AwayTeam"]]
  }
  
  colnames(dataFrame)[which(names(dataFrame) == "HomeTeam")] = "team1"
  colnames(dataFrame)[which(names(dataFrame) == "AwayTeam")] = "team2"
  colnames(dataFrame)[which(names(dataFrame) == "FTHG")] = "score1"
  colnames(dataFrame)[which(names(dataFrame) == "FTAG")] = "score2"
  
  # replace NA value to 0
  l = dim(dataFrame)[2]
  for (i in 1:l) {
    dataFrame[,i][is.na(dataFrame[,i])] = 0
  }
  
  return(dataFrame)
  
}

